/*
    Gn: A Server for the Internet Gopher Protocol(*).
    File: gn/init.c
    Version 1.0

    Copyright (C) 1993  <by John Franks>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    (*) Gopher is a registered trademark of the Univ. of Minn.
*/

/*  Improved logging support thanks to Paul Dubois:

Logging can be done either to a named file, or using syslog.  If a named
file is used, it's locked if possible.

Two complications:
- Not all systems have flock(), or at least not flock()'s that have the
struct flock used below.
- syslog() isn't universally available.

To get around this, there are two levels of logging selection/restriction.
- The first level occurs at compile time, and compiles code only for
those facilities that are available.  (syslog() code isn't compiled if
NO_SYSLOG is defined, and the flock() stuff isn't compiled if
NO_FLOCK is defined.)

- The second level occurs at runtime and allows selection from such facilities
as are available.  Note that -L file can be given on the command line, but
if flock() isn't available, it's possible for the log file to be corrupted
by two competing gn processes.  -S is recommended as an alternative.  If
syslog() is unavailable, specifying -S on the command line prints a warning
and no logging is done.  -L file is recommended as an alternative.
*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#if defined( __STDC__) && !defined( NeXT)
#include <unistd.h>
#endif
#include <fcntl.h>
#include "gn.h"
#ifndef NO_SYSLOG
#include <syslog.h>
#endif

/* this may be needed if unistd.h isn't included... */
#ifndef SEEK_SET
#define SEEK_SET 0
#endif

#ifdef NO_VOID_PTR
#define	void	int
#endif

extern char *optarg;
extern int optind;

char	rootdir[PATHLEN],
	root_title[PATHLEN],
	cfname[PATHLEN],
	afname[PATHLEN],
	remotehost[BUFSIZE],
	remaddr[PATHLEN];

int	accesstype = FREE;

#ifndef NO_FLOCK
static void	locklog(),
		unlocklog();
#endif

static void	open_gnlog(),
		log_logfile ();

static FILE	*logfp = NULL;

static char	logfile[PATHLEN];

#ifndef NO_SYSLOG
static void	log_syslog ();
#endif

static void	(*log_func) () = (void (*)()) NULL;

void
gn_init( argc, argv)
int	argc;
char	*argv[];
{
	int	c,
		errflg = 0;

/* Let's run as specified user and group if GID_SET defined.
 * This is imperitive if inetd cannot change UID from root.
 * (This has been tested with Ultrix 4.x.)
 */
#ifdef GID_SET
#ifndef UID_SET
#define UID_SET -2
#endif
 	setgid(  (gid_t) GID_SET );   
 	setegid( (gid_t) GID_SET );
 	seteuid( (uid_t) UID_SET );
 	setuid(  (uid_t) UID_SET );  /* This pulls rug out from under us */
#endif

	alarm( TIMEOUT);  /* Timeout after TIMEOUT seconds */
	strcpy( rootdir, ROOT_DIR);
	strcpy( root_title, ROOT_MENU_NAME);
	strcpy( cfname, CACHEFNAME);
	strcpy( afname, ACCESSFNAME);
	logfile[0] = '\0';

	while ((c = getopt(argc, argv, "aAL:c:g:t:S")) != -1) {
		switch ((char) c) {
			case 'a':
				accesstype = ROOTCHK;
				break;
			case 'A':
				accesstype = DIRCHK;
				break;
			case 'L':
				open_gnlog( optarg);
				log_func = log_logfile;
				break;
			case 'c':
				strcpy( cfname, optarg);
				break;
			case 'g':
				strcpy( afname, optarg);
				break;
			case 't':
				strcpy( root_title, optarg);
				break;
			case 'S':
#ifdef NO_SYSLOG
				senderr2( "-S option unavailable, no syslog",
						 "");
				exit( 2);
#else
				log_func = log_syslog;
#endif
				break;
			case '?':
				errflg++;
		}
	}

	if (errflg) {
		senderr2( "Usage: gn [-L logfile | -S] [-a | -A] [-c cache] ",
			"[-t title] [rootdir]\n");
		exit (2);
	}
	if ( argv[optind] )
		strcpy( rootdir, argv[optind]);

	get_remote_info( remaddr, remotehost);
	strlower( remotehost);

#ifndef NO_SYSLOG
	if (log_func == log_syslog)
	{

# ifdef LOG_DAEMON
		/* 4.3 style */
		openlog ("gn", LOG_PID | LOG_NDELAY, LOG_DAEMON);
# else
		/* 4.2 style */
		openlog ("gn", LOG_PID);
# endif
	}
#endif
}


static void
open_gnlog( logfile)
char	*logfile;
{
	if ( *logfile != '\0')
		if ( (logfp = fopen( logfile, "a")) == NULL ) {
			senderr2( "Can't open logfile: ", logfile);
			exit( 2);
		}
}

void
writelog( msg, msg2)
char	*msg,
	*msg2;
{
	if (log_func != (void (*)()) NULL)
		(*log_func) (msg, msg2);
}


static void
log_logfile( msg, msg2)
char	*msg,
	*msg2;
{
#ifndef NO_FLOCK
	struct flock	lck;
#endif

	time_t	clock;
	char	*cp,
		*ctime();

	time(&clock);
	cp = ctime(&clock);

#ifndef NO_FLOCK
	locklog( &lck, fileno(logfp) );
#endif
	fseek( logfp, 0L, 2);   
	fprintf(logfp, "%s: %s (%s): %s",  remotehost, msg, msg2, cp);

#ifndef NO_FLOCK
	unlocklog( &lck, fileno(logfp) );   
#endif
}

#ifndef NO_FLOCK
static void
locklog(lck, fd)
struct flock	*lck;
int	fd;
{
	lck->l_type = F_WRLCK;
	lck->l_whence = SEEK_SET;
	lck->l_start = 0L;
	lck->l_len = 0L;
	fcntl(fd, F_SETLKW, lck);
}

static void
unlocklog( lck, fd)
struct flock	*lck;
int	fd;
{
	lck->l_type = F_UNLCK;
	fcntl( fd, F_SETLKW, lck);
}
#endif


#ifndef NO_SYSLOG

/*
 * syslog() provides the time automatically, no need to put it in the message.
 */

static void
log_syslog( msg, msg2)
char	*msg,
	*msg2;
{
	syslog(LOG_INFO, "%s: %s (%s)",  remotehost, msg, msg2);
}
#endif

