/*
 * httpd.c: simple http daemon for answering WWW file requests
 *
 * 
 * Rob McCool 3/21/93
 * 
 */


#include "httpd.h"

void usage(char *bin) {
    fprintf(stderr,"Usage: %s [-d directory]\n",bin);
    fprintf(stderr,"-d directory : specify an alternate config file\n");
    exit(1);
}

int sd;
pid_t pgrp;

void detach() {
    int x;

    chdir("/");
    if((x = fork()) > 0)
        exit(0);
    else if(x == -1) {
        fprintf(stderr,"httpd: unable to fork new process\n");
        perror("fork");
        exit(1);
    }
#ifndef NO_SETSID
    if((pgrp=setsid()) == -1) {
        fprintf(stderr,"httpd: setsid failed\n");
        perror("setsid");
        exit(1);
    }
#else
    if((pgrp=setpgrp(getpid(),0)) == -1) {
        fprintf(stderr,"httpd: setpgrp failed\n");
        perror("setpgrp");
        exit(1);
    }
#endif    
}

void sig_term() {
    log_error("httpd: caught SIGTERM, shutting down");
#ifndef NO_KILLPG
    killpg(pgrp,SIGKILL);
#else
    kill(-pgrp,SIGKILL);
#endif
    close(sd);
}

#ifdef BSD
void ign() {
#ifndef NEXT
    int status;
#else
    union wait status;
#endif
    pid_t pid;

    while( (pid = wait3(&status, WNOHANG, NULL)) > 0);
}
#endif

void bus_error() {
    log_error("httpd: caught SIGBUS, dumping core");
    chdir(server_root);
    abort();         
}

void seg_fault() {
    log_error("httpd: caught SIGSEGV, dumping core");
    chdir(server_root);
    abort();
}

void set_signals();

void restart() {
    log_error("httpd: caught SIGHUP, restarting");
    close_logs();
    kill_mime();
    kill_security();
    if(server_hostname) {
        free(server_hostname);
        server_hostname = NULL;
    }
    read_config();
    open_logs();
    get_local_host();
    set_signals();
    log_error("httpd: successful restart");
}

void set_signals() {
    signal(SIGSEGV,(void (*)())seg_fault);
    signal(SIGBUS,(void (*)())bus_error);
    signal(SIGTERM,(void (*)())sig_term);
    signal(SIGHUP,(void (*)())restart);

#ifdef BSD
    signal(SIGCHLD,(void (*)())ign);
#else
    signal(SIGCHLD,SIG_IGN);
#endif
}


void standalone_main() {
    int csd, clen,pid, one=1;
    struct sockaddr_in sa_server,sa_client;

    detach();

    if ((sd = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP)) == -1) {
        fprintf(stderr,"httpd: could not get socket\n");
        perror("socket");
        exit(1);
    }

    if((setsockopt(sd,SOL_SOCKET,SO_REUSEADDR,&one,sizeof(one))) == -1) {
        fprintf(stderr,"httpd: could not set socket option\n");
        perror("setsockopt");
        exit(1);
    }
    bzero((char *) &sa_server, sizeof(sa_server));
    sa_server.sin_family=AF_INET;
    sa_server.sin_addr.s_addr=htonl(INADDR_ANY);
    sa_server.sin_port=htons(port);
    if(bind(sd,(struct sockaddr *) &sa_server,sizeof(sa_server)) == -1) {
        fprintf(stderr,"httpd: could not bind to port %d\n",port);
        perror("bind");
        exit(1);
    }
    listen(sd,5);

    set_signals();
    log_pid();

    while(1) {
      retry:
        clen=sizeof(sa_client);
        if((csd=accept(sd,&sa_client,&clen)) == -1) {
            if(errno == EINTR)  {
#ifdef BSD
                ign();
#endif
                goto retry;
            }
            log_error("socket error: accept failed");
        }
	/* we do this here so that logs can be opened as root */
        if((pid = fork()) == -1)
            log_error("unable to fork new process");
        else if(!pid) {
	    struct passwd* pwent;

            /* Only effective if we're running as root */
            if(!getuid()) {
                /* Now, make absolutely certain we don't have any privileges
                 * except those mentioned in the configuration file. */
                if ((pwent = getpwuid(user_id)) == NULL) {
                    log_error("couldn't determine user name from uid");
                    exit(-1);
                }
                /* Reset `groups' attribute. */
                if (initgroups(pwent->pw_name, group_id) == -1) {
                    log_error("unable to setgroups");
                    exit(-1);
                }
                /* Note the order, first setgid() and then setuid(), it
                 * wouldn't work the other way around. */
                if (setgid(group_id) == -1) {
                    log_error("unable change gid");
                    exit(-1);
                }
                if (setuid(user_id) == -1) {
                    log_error("unable change uid");
                    exit(-1);
                }
            }
            close(0);
            close(1);
            dup2(csd,0);
            dup2(csd,1);
            close(sd);
            process_request(stdin,stdout);
            fclose(stdin);
            fclose(stdout);
            exit(0);
        }
            close(csd);
        }
    }

extern char *optarg;
extern int optind;

main(int argc, char *argv[])
{
    int c;

    strcpy(server_root,HTTPD_ROOT);
    while((c = getopt(argc,argv,"d:")) != -1) {
        switch(c) {
          case 'd':
            strcpy(server_root,optarg);
            break;
          case '?':
            usage(argv[0]);
        }
    }

    read_config();
    open_logs();
    get_local_host();

    set_env_vars();
    if(standalone)
        standalone_main();
    else {
        user_id = getuid();
        group_id = getgid();
        port = get_portnum(fileno(stdout),stdout);
        process_request(stdin,stdout);
    }
    fclose(stdin);
    fclose(stdout);
    exit(0);
}
